==> mars-0.9.2.tar.gz: mars-0.9.2/src/mpu/lib/mars_task_semaphore.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <stdlib.h>
#include <string.h>

#include "mars/mars_task_semaphore.h"
#include "mars/mars_task_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_semaphore semaphore;

int mars_task_semaphore_initialize(uint64_t semaphore_ea, int32_t count)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(abs(count) <= MARS_TASK_SEMAPHORE_WAIT_MAX,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(semaphore_ea);
	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	semaphore.mars_context_ea = mars_get_mars_context_ea();
	semaphore.count = count;
	semaphore.wait_count = 0;

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}

int mars_task_semaphore_acquire(uint64_t semaphore_ea)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	struct mars_task_context *task;

	task = (struct mars_task_context *)mars_get_workload_context();

	/* make sure task context has a context save area */
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	/* check if semaphore wait limit reached */
	MARS_CHECK_CLEANUP_RET(
		semaphore.wait_count < MARS_TASK_SEMAPHORE_WAIT_MAX,
		mars_mutex_unlock_put(semaphore_ea,
				(struct mars_mutex *)&semaphore),
		MARS_ERROR_LIMIT);

	if (semaphore.count <= 0) {
		/* add id to wait list */
		semaphore.wait_id[semaphore.wait_count] =
			mars_get_workload_id();
		semaphore.wait_count++;

		mars_mutex_unlock_put(semaphore_ea,
				(struct mars_mutex *)&semaphore);

		/* wait for signal */
		mars_signal_wait();

		return MARS_SUCCESS;
	}

	/* decrement semaphore count */
	semaphore.count--;

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}

int mars_task_semaphore_release(uint64_t semaphore_ea)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	/* increment semaphore count */
	semaphore.count++;

	/* signal those that are waiting */
	if (semaphore.count > 0 && semaphore.wait_count) {
		/* decrement semaphore count */
		semaphore.count--;

		/* signal waiting task */
		mars_signal_send(semaphore.wait_id[0]);

		/* flush id from wait list */
		semaphore.wait_count--;
		memmove(&semaphore.wait_id[0],
			&semaphore.wait_id[1],
			sizeof(uint16_t) * semaphore.wait_count);
	}

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/mpu/lib/mars_task_semaphore.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <stdlib.h>
#include <string.h>

#include "mars/mars_task_semaphore.h"
#include "mars/mars_workload_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_semaphore semaphore;

int mars_task_semaphore_initialize(uint64_t semaphore_ea, int32_t count)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(abs(count) <= MARS_TASK_SEMAPHORE_WAIT_MAX,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(semaphore_ea);
	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	semaphore.mars_context_ea = mars_get_mars_context_ea();
	semaphore.count = count;
	semaphore.wait_count = 0;

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}

int mars_task_semaphore_acquire(uint64_t semaphore_ea)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	struct mars_task_context *task;

	/* make sure task context has a context save area */
	task = &mars_get_workload_context()->type_context.task;
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	/* check if semaphore wait limit reached */
	MARS_CHECK_CLEANUP_RET(
		semaphore.wait_count < MARS_TASK_SEMAPHORE_WAIT_MAX,
		mars_mutex_unlock_put(semaphore_ea,
				(struct mars_mutex *)&semaphore),
		MARS_ERROR_LIMIT);

	if (semaphore.count <= 0) {
		/* add id to wait list */
		semaphore.wait_id[semaphore.wait_count] =
			mars_get_workload_id();
		semaphore.wait_count++;

		mars_mutex_unlock_put(semaphore_ea,
				(struct mars_mutex *)&semaphore);

		/* wait for signal */
		mars_signal_wait();

		return MARS_SUCCESS;
	}

	/* decrement semaphore count */
	semaphore.count--;

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}

int mars_task_semaphore_release(uint64_t semaphore_ea)
{
	MARS_CHECK_RET(semaphore_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((semaphore_ea & MARS_TASK_SEMAPHORE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(semaphore_ea, (struct mars_mutex *)&semaphore);

	/* increment semaphore count */
	semaphore.count++;

	/* signal those that are waiting */
	if (semaphore.count > 0 && semaphore.wait_count) {
		/* decrement semaphore count */
		semaphore.count--;

		/* signal waiting task */
		mars_signal_send(semaphore.wait_id[0]);

		/* flush id from wait list */
		semaphore.wait_count--;
		memmove(&semaphore.wait_id[0],
			&semaphore.wait_id[1],
			sizeof(uint16_t) * semaphore.wait_count);
	}

	mars_mutex_unlock_put(semaphore_ea, (struct mars_mutex *)&semaphore);

	return MARS_SUCCESS;
}
